from typing import Optional

from spot_llm.models.hf_model import HFSpotter


class Opt125Spotter(HFSpotter):
    def __init__(self, token: Optional[str] = None) -> None:
        super().__init__(model_id="facebook/opt-125m", token=token)



from typing import Optional

import torch
from torch import Tensor
from transformers import AutoTokenizer, AutoModelForCausalLM  # type: ignore

from spot_llm.models.spotter import Spotter


class HFSpotter(Spotter):
    def __init__(self, model_id: str, token: Optional[str] = None):
        torch.set_default_dtype(torch.float16)
        super().__init__(model_id=model_id, token=token)

    def initialize(self) -> None:
        self.tokenizer = AutoTokenizer.from_pretrained(self.model_id, token=self.token)
        self.model = AutoModelForCausalLM.from_pretrained(
            self.model_id, token=self.token, device_map="auto"
        )

    def tokenize(self, text: str) -> Tensor:
        tokenz = self.tokenizer(
            text, add_special_tokens=False, return_tensors="pt"
        ).input_ids
        return tokenz

    def _get_lazy_logits(
        self,
        inputs: Tensor,
    ) -> Tensor:
        with torch.no_grad():
            output = self.model(inputs.to(self.model.device))
        return output



from spot_llm.models.hf_model import HFSpotter


class MistralSpotter(HFSpotter):
    def __init__(self, token: str, version: str = "0.3"):
        super().__init__(model_id=f"mistralai/Mistral-7B-v{version}", token=token)



from abc import abstractmethod
from typing import Any, List, Optional, Tuple

import numpy as np
from torch import Tensor


class Spotter:
    def __init__(
        self,
        model_id: str,
        token: Optional[str] = None,
        base_token_count: int = 10,
    ) -> None:
        self.model_id = model_id
        self.token = token
        self.base_token_count = base_token_count
        self.tokenizer: Any
        self.model: Any
        self.initialize()

    @abstractmethod
    def initialize(self) -> None:
        raise NotImplementedError()

    def predict_logits_and_tokenz(self, input_prompt: str) -> Tuple[Tensor, Tensor]:
        tokenz = self.tokenize(input_prompt)
        return self._get_lazy_logits(tokenz), tokenz

    @abstractmethod
    def tokenize(self, text: str) -> Tensor:
        raise NotImplementedError()

    @abstractmethod
    def _get_lazy_logits(
        self,
        inputs: Tensor,
    ) -> Tensor:
        raise NotImplementedError()

    def score(self, text: str) -> float:

        outputs, tokenz = self.predict_logits_and_tokenz(text)

        output = outputs[0][0]
        rankings: List[int] = []
        for i in range(output.shape[0] - 1):
            if i < self.base_token_count:
                continue
            logit = output[i - 1, :]
            actual_token = tokenz[0, i]
            rankings.append(int((logit > logit[actual_token]).sum().item()))

        if len(rankings) == 0:
            return -1.0
        return float(np.mean(rankings) / output.shape[1])

    def is_human(self, text: str, treshold: float = 0.1) -> bool:
        return self.score(text) >= treshold
